"""
Robot AI

  General design:

  Choose best of a few random programs for each AI robot.
  based on over-simplistic rules.

  Fill in likely locations of AI robots as their program
  is chosen.

  Do faster/cleverer robots first, in same order as
  the RoboFight will sort them.


"""
import pygame

from threading import Thread
from random import choice, randint
from math import cos,sin,radians

def direction(angle):
    # dx, dy as nearest NSEW direction to angle
    dxf = cos(radians(angle))
    dyf = sin(radians(angle))
    # round to nearest int
    return int(round(dxf)), int(round(dyf))

def distance(xyz0, xyz1):
    x0,y0,z0 = xyz0
    x1,y1,z1 = xyz1
    return ((x0-x1)**2 + (y0-y1)**2 + (z0-z1)**2)**0.5

class RobotAIThread(Thread):
    def __init__(self,stackmap,robots,team):
        Thread.__init__(self,name="RobotAI for " + team)
        self.pmap = {}
        self.stackmap = stackmap
        self.ai_team = team
        self.sorted_robots = sorted(
            robots, key = lambda r: -(r.speed + r.clever))
        self.interrupted = False
        self.running = False
    def random_program(self,r):
        moves = ["Left","Right","Walk","Walk","Walk"] # bias for walking
        default_attack = "Attack"
        if r["armr"]:
            moves.append("Attack")
        if r["arml"]:
            moves.append("AttackL")
            default_attack = "AttackL"
        prog = []
        max_steps = r.moves
        max_attacks = r.attacks
        attacks = 0
        plength = randint(max_steps//2,max_steps)
        while len(prog) < plength:
            move = choice(moves)
            if move.startswith("Attack"):
                if attacks< max_attacks:
                    attacks += 1
                else:
                    continue
            prog.append(move)
        if "Attack" not in prog and "AttackL" not in prog:
            if len(prog) == r.moves:
                prog[-1:] = [default_attack]
            else:
                prog.append(default_attack)
        return prog
    def interrupt(self):
        self.interrupted = True
    def mapget(self,x,y,t,ai_team):
        get = self.pmap.get
        for t1 in range(t,-1,-1):
            p = get((x,t,t1,ai_team),0)
            if p or not t1:
                return p
    def run(self):
        # give AI team robots random progs
        for r in self.sorted_robots:
            if r.team == self.ai_team:
                x,y,z = r.pos
                x = int(round(x))
                y = int(round(y))
                r.program = self.random_program(r)
                self.pmap[x,y,0,True] = 1.0
        print "<AI> Assigned random robot programs"
        # Get opponent positions
        targets = [r.pos for r in self.sorted_robots
                   if r.team != self.ai_team]
        # Choose best random program for each robot
        for r in self.sorted_robots:
            if r.team != self.ai_team: continue
            progs = [self.random_program(r)
                     for i in range(r.moves ** 2 + 50)]
            bestprog = None
            bestscore = -1000
            besttrack = None
            x,y,rz = r.pos
            rx = int(round(x))
            ry = int(round(y))
            rdx,rdy = direction(r.angle)
            mapget = self.mapget
            if targets:
                my_target = sorted(targets,
                               key = lambda pos: distance(r.pos,pos)
                               )[0]
            else:
                my_target = None
            for prog in progs:
                positions = []
                certainty = 1.0
                x,y,z = rx,ry,rz
                dx,dy = rdx,rdy
                score = 0
                for t,move in enumerate(prog):
                    if move == "Left":
                        dy,dx = dx, -dy
                    elif move == "Right":
                        dx,dy = dy, -dx
                    elif move == "Walk":
                        h = self.stackmap.floor_height(x+dx,y+dy)
                        if h is None or abs(h - z) > r.jump:
                            pass
                        else:
                            x, y, z = x+dx, y+dy, h
                    elif move.startswith("Attack"):
                        h = self.stackmap.floor_height(x+dx,y+dy)
                        if h is None or abs(h - z) > r.reach:
                            pass
                        else:
                            prob_hit_ally = self.mapget(x+dx,y+dy,t,True)
                            score -= prob_hit_ally * certainty
                    prob_obstruct_ally = mapget(x,y,t,True)
                    score -= prob_obstruct_ally * certainty
                    certainty *= (1.0 - min(prob_obstruct_ally,1.0))
                    positions.append((x,y,certainty))
                # Try to be nearer target than before
                if my_target:
                    score += (distance(r.pos, my_target) -
                              distance((x,y,z),my_target))
                if score > bestscore:
                    bestprog = prog
                    bestscore = score
                    besttrack = positions
            r.program = bestprog
            print "<AI> chose program for", r._name,"(score =",bestscore,")"
#            print "<AI debug>",bestprog
            for t, (x,y,c) in enumerate(besttrack):
                prob = mapget(x,y,t,True)
                self.pmap[x,y,t,True] = max(prob,c)
                
            # Are we done?
            if self.interrupted:
                break
