"""
  Stacks of blocks
"""
from tdgl import resource, part, objpart, picking
from OpenGL.GL import *

__all__ = ['Block','Item','Stack','StackMap']

class Block(object):
    is_floor = True
    height = 0
    is_obstacle = False
    is_item = False
    is_breakable = False
    obj = None
    def __init__(self,height,filename,pieces=(),angle=0):
        self.height = height
        self.obj = objpart.get_obj(filename)
        self.pieces = pieces
        if not pieces:
            self.pieces = self.obj.pieces()
        self.angle = angle
    def render(self,mode):
        if mode == "PICK":
            picking.label((self.__class__.__name__,self))
        if self.angle:
            glRotate(self.angle,0,0,1)
        for p in self.pieces:
            self.obj.drawpiece(p)
        if mode == "PICK":
            picking.nolabel()

class Obstacle(Block):
    is_floor = False
    is_obstacle = True

class Button(Obstacle):
    on = False
    def render(self,mode):
        if mode == "PICK":
            picking.label((self.__class__.__name__,self))
        if self.angle:
            glRotate(self.angle,0,0,1)
        self.obj.drawpiece("Base")
        mtl = "DarkRed"
        if self.on:
            mtl = "BrightRed"
            glTranslate(0,0,-0.1)
        self.obj.mat_dls.select(mtl)
        self.obj.drawpiece("Button")
        if mode == "PICK":
            picking.nolabel()

class Item(Block):
    is_floor = False
    is_item = True
    
class Stack(object):
    def __init__(self,blocks=[]):
        self.blocks = list(blocks)
        self.angle = 0.0

    def floor_height(self):
        """height of top walkable block, or None if obstructed"""
        h = 0
        for b in self.blocks:
            if b.is_floor:
                h += b.height
            else:
                if b.is_obstacle:
                    h = None
                break
        return h
    def item(self):
        """If topmost block is an item, that.  Otherwise None"""
        if self.blocks:
            top = self.blocks[-1]
            if top.is_item:
                return top
        return None
    def add(self,block):
        self.blocks.append(block)
    def add_under(self,block):
        self.blocks.insert(0,block)
    def rotate(self):
        glRotatef(90,0.0,0.0,1.0)
    def remove(self):
        self.blocks[-1:] = []
    def render(self,mode):
        glPushMatrix()
        for b in self.blocks:
            b.render(mode)
            glTranslate(0,0,b.height)
        glPopMatrix()

class StackMap(part.Part):
    def __init__(self,length,width,*args,**kw):
        super(StackMap,self).__init__("",*args,**kw)
        self.stacks = {}
        self.length = length
        self.width = width

    def floor_height(self,x,y):
        s = self.stacks.get((x,y))
        if s:
            return s.floor_height()
        else:
            return None

    def item(self,x,y):
        s = self.stacks.get((x,y))
        if s:
            return s.item()
        else:
            return None

    def add(self,x,y,symbols,mapping):
        """Call this with a list or string of symbols, each identifying a
        block in the mapping dictionary.  Those blocks will be added onto
        the stack at x,y"""
        stack = self.stacks.get((x,y))
        if not stack:
            stack = self.stacks[x,y] = Stack()
        for symbol in symbols:
            b = mapping.get(symbol)
            if b:
                stack.add(b)

    def add_under(self,x,y,symbol,mapping):
        stack = self.stacks.get((x,y))
        if not stack:
            stack = self.stacks[x,y] = Stack()
        b = mapping.get(symbol)
        if b:
            stack.add_under(b)

    def take(self,x,y):
        s = self.stacks.get((x,y))
        i = None
        if s:
            i = s.item()
            if i:
                s.remove()
        return i

    def setup_geom(self):
        glPushMatrix()
#        glTranslate(-0.5, -0.5,0)
    def setdown_geom(self):
        glPopMatrix()

    def render(self,mode):
        for (x,y), s in self.stacks.items():
            glPushMatrix()
            glTranslate(x,y,0)
            s.render(mode)
            glPopMatrix()

            
def unit_test():
    b1 = Block()
    b2 = Block()
    i1 = Item()
    b1.height = 1
    b2.height = 2
    sm = StackMap()
    mapping = {"1":b1, "2":b2, "i":i1}
    sm.add(1,1,"21i",mapping)
    sm.add(1,2,"2",mapping)
    sm.add(2,1,"",mapping)
    assert sm.floor_height(1,1) == 3
    assert sm.item(1,1) == i1
    assert sm.floor_height(1,2) == 2
    assert sm.item(1,2) == None
    assert sm.floor_height(2,1) == 0
    assert sm.item(2,1) == None
    assert sm.floor_height(2,2) == None
    assert sm.item(2,2) == None
    assert sm.take(1,2) == None
    assert sm.take(1,1) == i1
    assert sm.take(1,1) == None
    return True
