#! /usr/bin/env python

"""A Part is a thing that can be drawn or respond to events.
   A Group is a Part made of Parts.

   Parts have a certain amount of their configuration calculated
   from a stylesheet. They can be requested to refresh this information
   in the event of a stylesheet change.
   
   They also have a step() method to make changes for the next
   animation frame.

This file copyright Peter Harris 2004, released under the terms of the
GNU GPL. See www.gnu.org for details.
"""
from OpenGL.GL import glPushMatrix, glPopMatrix, glTranslate, glRotate, glScalef
from OpenGL.GL import glEnable, GL_NORMALIZE

import stylesheet, picking
try:
    set
except NameError:
    from sets import Set as set

def diffkeys(d1,d2):
    """Keys of d2 that are not in d1 or which have different values in d2"""
    s = set()
    for k,v in d2.items():
        if k not in d1 or d1[k] != v:
            s.add(k)
    return s

class Part(object):
    """A component of a scene.  It can draw itself, restyle itself according
    to a stylesheet, respond to an event, step to the next animation frame.
    """
    _default_geom = {}
    _default_style = {}
    _style_attributes = ()
    _expired = False
    _has_transparent = False # by default, won't need a transparency rendering pass
    _name = ""    # ID for use in stylesheets
    _visible = True
    _pickable = False
    _active = True  # relevant to whether a Group containing the part expires
    def __init__(self,name="",geom=None,style=None,**kw):
        super(Part,self).__init__(name,geom=geom,style=style,**kw)
        self._geom = self._default_geom.copy()
        if geom:
            self._geom.update(geom)
        self._style = self._default_style.copy()
        if style:
            self._style.update(style)
        if name:
            self._name = name
        self._style_classes = []

    def getgeom(self,name,default=None):
        return self._geom.setdefault(name,default)
    def getstyle(self,name,default=None):
        return self._style.setdefault(name,default)
    def setgeom(self,name,val):
        self._geom[name]=val
    def setstyle(self,name,val):
        self._style[name]=val

    def style_selector(self):
        """A selector for this object to use for stylesheet lookups"""
        sl = []
        if self._name:
            sl.append("#%s" % self._name)
        sl.append(self.__class__.__name__)
        sl.extend(self._style_classes)
        return '.'.join(sl)
        
    def restyle(self,force=False):
        """Re-calculate style information, and if anything has changed,
        prepare whatever expensive calculations, display lists or
        cached objects are needed for drawing with the new style.
        """
        style = stylesheet.get(self.style_selector(),*self._style_attributes)
        d = diffkeys(self._style,style)
        self._style.update(style)
        if d or force:
            self.prepare()
            
    def prepare(self):
        """Perform expensive calculations, load textures or fonts etc.
        Anything that you don't want to be doing every single frame."""
        pass
        
    def draw(self,mode):
        """Draw, in one of three ways used in different contexts:
          OPAQUE      : draw the opaque bits, with styling
          TRANSPARENT : draw the transparent bits, with styling
          PICK        : draw both but with no styling, and set picking labels
          
          A rendering loop will generally call draw("OPAQUE") for all objects,
          then draw("TRANSPARENT").  When picking is underway, it will call
          draw("PICK").
          
          NB: if you can be sure no pixels will be drawn with alpha < 1.0, you can
          skip the TRANSPARENT pass. Set _has_transparent to False. 
        """
        if not self._visible: return    # turning off _visible turns off draw()
        #print "drawing", self.__class__.__name__, self._name, mode #DEBUG
        self.setup_geom()   # position, scaling, rotation etc.
        if mode != "PICK":  self.setup_style()
        if self._has_transparent or mode != 'TRANSPARENT':
            self.render(mode)
        if mode != "PICK":  self.setdown_style()
        self.setdown_geom()
        
    def setup_geom(self):
        """Default setup_geom is translation to pos, rotation by angle"""
        pos = self.pos
        angle = self.angle
        glPushMatrix()
        glTranslate(*pos)
        if angle:
            glRotate(angle,0,0,1.0)
        
    def setup_style(self): pass
    
    def render(self,mode): pass
    
    def setdown_style(self): pass
    
    def setdown_geom(self):
        """Default setdown_geom pops modelview matrix"""
        glPopMatrix()
        
    def event(self,event): pass
    
    def step(self,ms=20): pass
    
    def expired(self):
        """Return a true value if there is nothing to draw any more"""
        return self._expired
        
    # pos and angle properties
    def _getpos(self):
        return self._geom.get('pos',(0,0,0))
    def _setpos(self,pos):
        self._geom['pos'] = pos
    pos = property(_getpos,_setpos)
    def _getangle(self):
        return self._geom.get('angle',0.0)
    def _setangle(self,angle):
        self._geom['angle'] = angle
    angle = property(_getangle,_setangle)

class ScalePart(Part):
    def setup_geom(self):
        """Move to pos, turn by angle and scale by scale"""
        super(ScalePart,self).setup_geom()
        scale=self._geom.get('scale',1)
        if scale and scale != 1:
            glEnable(GL_NORMALIZE)
            glScalef(scale,scale,scale)

class Group(Part):
    """A collection of parts, bundled together for convenience.
    The group's setup_*() and setdown_*() methods bracket the drawing
    of all drawable parts.
    All parts that can respond to events are passed each event passed
    into the event() method.
    All parts that have a name are kept in a dictionary so they can be
    found easily."""
    _transient = False
    _has_transparent = True # contents might need a transparency rendering pass
    def __init__(self,name,contents,**kwd):
        super(Group,self).__init__(name,**kwd)
        self._transient = kwd.get('transient',False)
        self.named_parts = {}
        self.contents = []
        for p in contents:
            self.append(p)
    def render(self,mode):
        for p in self.contents:
            p.draw(mode)
    def restyle(self,force=False):
        super(Group,self).restyle(force)
        for p in self.contents:
            p.restyle(force)
    def event(self,event):
        return self.event_to_contents(event)
    def event_to_contents(self,event):
        """Call the event() method of every Part in the contents
        return first non-None return value (indicating some Part
        has consumed the event)
        """
        res = []
        for p in self.contents:
            r = p.event(event)
            if r is not None:
                return r
    def step(self,ms=20):
        self.step_contents(ms)
    def step_contents(self,ms=20):
        """Next animation frame of each visible part, then remove
        any that have expired"""
        any_active = False
        for p in tuple(self.contents):
            if p.expired():
                self.remove(p)
            else:
                p.step(ms)
                any_active = any_active or p._active
        if self._transient and not any_active:
            self._expired = True
    def append(self,newpart):
        """Add another part to the contents of the group"""
        self.contents.append(newpart)
        if newpart._name:
            self.named_parts[newpart._name] = newpart
    def remove(self,delpart):
        """Remove a part if it is in the contents of a group"""
        try:
            self.contents.remove(delpart)
        except ValueError:
            pass
        name = delpart._name
        if (name in self.named_parts and
           self.named_parts[name] == delpart):
            del self.named_parts[name]
    def __getitem__(self,name):
        """Get a named part from contents of group, or from any group
        within it"""
        try:
            return self.named_parts[name]
        except KeyError:
            for g in self.contents:
                if isinstance(g,Group):
                    p = g[name]
                    if p:
                        return p
            return None
